import typing as t

from pyarrow.flight import ActionType, FlightClient, Ticket  # type: ignore
from pydantic import BaseModel

from tktl.core.serializers import deserialize_arrow

from .utils import batch_arrow


class EndpointActionGroup(BaseModel):
    endpoint: ActionType
    X: ActionType
    y: ActionType


class SDKArrowEndpoint:
    """SDKArrowEndpoint.

    This is a SDK Arrow endpoint, exposing function calling through
    `__call__` and sample data through `X` and `y`.
    """

    def __init__(self, *, client: FlightClient, action_group: EndpointActionGroup):
        self._client = client
        self._action_group = action_group
        self.__name__ = action_group.endpoint.type

    def __call__(self, X: t.Any) -> t.Any:
        return batch_arrow(client=self._client, action=self._action_group.endpoint, X=X)

    def X(self) -> t.Any:
        reader = self._client.do_get(Ticket(ticket=self._action_group.X.type))
        return deserialize_arrow(reader.read_all())

    def y(self) -> t.Any:
        reader = self._client.do_get(Ticket(ticket=self._action_group.y.type))
        return deserialize_arrow(reader.read_all())


class ArrowEndpoints:
    """ArrowEndpoints.

    This is the `endpoints` object on TaktileSDK Arrow Clients.
    """

    def __init__(self, *, client: FlightClient, actions: t.List[ActionType]):

        action_types = [a.type for a in actions]
        action_groups = []

        for action in actions:

            if (
                action.type + "__X" in action_types
                and action.type + "__y" in action_types
            ):
                action_groups.append(
                    EndpointActionGroup(
                        endpoint=action,
                        X=[a for a in actions if a.type == action.type + "__X"][0],
                        y=[a for a in actions if a.type == action.type + "__y"][0],
                    )
                )

        for group in action_groups:
            setattr(
                self,
                group.endpoint.type,
                SDKArrowEndpoint(client=client, action_group=group),
            )



import json
import os
import pathlib

import pytest

from tktl.commands.login import LogInCommand, SetApiKeyCommand
from tktl.core.exceptions import APIClientException


def test_set_api_key_command(capsys):
    cmd = SetApiKeyCommand()
    cmd.execute(api_key=None)
    out, err = capsys.readouterr()
    assert "API Key cannot be empty.\n" == err

    cmd.execute(api_key="ABC")
    config_path = pathlib.Path.home() / ".config" / "tktl" / "config.json"
    assert config_path.exists()
    with open(config_path, "r") as j:
        d = json.load(j)
        assert d["api-key"] == "ABC"


@pytest.mark.skipif("TEST_USER_API_KEY" not in os.environ, reason="No User Logged In")
def test_login_command(logged_in_context, capsys):
    cmd = LogInCommand()
    assert cmd.execute() is True
    out, err = capsys.readouterr()
    assert out == f"Authentication successful for user: {os.environ['TEST_USER']}\n"


def test_login_fail_command(capfd):
    cmd = LogInCommand()
    with pytest.raises(APIClientException):
        cmd.execute()



from __future__ import annotations

import typing as t

import numpy as np  # type: ignore
import pandas as pd  # type: ignore

from tktl.core import ExtendedEnum


class ArrowFormatKinds(str, ExtendedEnum):
    DATAFRAME = "dataframe"
    SERIES = "series"
    ARRAY = "array"

    @staticmethod
    def from_type(t: t.Union[pd.DataFrame, pd.Series, np.ndarray]) -> ArrowFormatKinds:
        if isinstance(t, pd.DataFrame):
            return ArrowFormatKinds.DATAFRAME
        if isinstance(t, pd.Series):
            return ArrowFormatKinds.SERIES
        if isinstance(t, np.ndarray):
            return ArrowFormatKinds.ARRAY
        raise ValueError(f"{type(t)} is not an arrow communication type")



from fastapi import BackgroundTasks  # noqa: F401
from taktile_client import ArrowClient, RestClient  # noqa: F401

from .registration.decorator import Tktl  # noqa: F401
from .registration.endpoints import (  # noqa: F401
    ArrowEndpoint,
    EndpointResponse,
    ProfiledEndpoint,
    TypedEndpoint,
)
from .settings import settings  # noqa: F401



import click

from tktl.cli.common import get_secrets
from tktl.commands.run import run_container


@click.command()
@click.option(
    "--path",
    "-p",
    help="Directory where the deployment lives",
    required=False,
    default=".",
)
@click.option(
    "--secrets-repository",
    "-s",
    help="Full repository name (owner/name) to use from which to get the secret names",
    required=False,
)
@click.option(
    "--cache/--disable-cache",
    help="Enable or disable using cache of images",
    is_flag=True,
    default=True,
)
@click.option(
    "--prune/--disable-prune",
    help="Enable or disable automatic pruning of images",
    is_flag=True,
    default=False,
)
@click.option("--background", help="Run the image in the background", is_flag=True)
@click.option(
    "--auth/--no-auth",
    help="Run with auth enabled or disabled. Enabled by default",
    default=True,
)
@click.option(
    "--reload/--no-reload",
    help="Run with hot-reloading for REST. Enabled by default",
    default=True,
)
@click.option(
    "--color/--no-color",
    help="Enable or disable colored output",
    is_flag=True,
    default=True,
)
def run(
    path: str,
    cache: bool,
    prune: bool,
    background: bool,
    auth: bool,
    reload: bool,
    color: bool,
    secrets_repository: str,
) -> None:
    """
    Run a deployment locally
    """
    secrets = get_secrets(secrets_repository=secrets_repository)
    run_container(
        path=path,
        cache=cache,
        prune=prune,
        background=background,
        auth_enabled=auth,
        reload=reload,
        color_logs=color,
        secrets=secrets,
    )



from requests.exceptions import HTTPError

from tktl import __version__
from tktl.commands import BaseTaktileApiCommand, CommandBase
from tktl.core.config import set_api_key, settings
from tktl.core.exceptions import APIClientException
from tktl.core.loggers import LOG
from tktl.core.schemas.user import TaktileUser
from tktl.login import logout
from tktl.version import TaktileVersionChecker


class LogInCommand(BaseTaktileApiCommand):
    def execute(self):
        try:
            user = self.client.call(
                verb="get",
                path=f"{settings.API_V1_STR}/users/me",
                model=TaktileUser,
            )
        except HTTPError as e:
            raise APIClientException(e.response.status_code, e.response.reason)
        LOG.log(f"Authentication successful for user: {user.username}", color="green")
        return True


class LogOutCommand(CommandBase):
    def execute(self):
        logout()


class ShowVersionCommand(CommandBase):
    def execute(self, check: bool = False):
        LOG.log(__version__)
        if check:
            TaktileVersionChecker.look_for_new_version()


class SetApiKeyCommand(CommandBase):
    def execute(self, api_key):
        if not api_key:
            LOG.error("API Key cannot be empty.")
            return
        set_api_key(api_key)
        LOG.log(
            f"Successfully added your API Key to {settings.TAKTILE_CONFIG_FILE} You're ready to go!"
        )



import pytest
from src.endpoints import client
from taktile_types.enums.endpoint import EndpointKinds


@pytest.mark.parametrize(
    "endpoint",
    [
        endpoint
        for endpoint in client.endpoints
        if endpoint.kind in [EndpointKinds.ARROW, EndpointKinds.PROFILED]
    ],
)
def test_endpoints(json_metadata, endpoint):
    """test_endpoints.
    This test ensures the provided sample data on endpoints can be correctly
    processed by the endpoints.

    It is recommended to keep these tests around.
    """

    json_metadata["section"] = "Taktile Automatic Endpoint Tests"
    json_metadata["pass_message"] = f"Sample data for {endpoint.name} is valid"
    json_metadata["fail_message"] = f"Sample data for {endpoint.name} is invalid"

    endpoint.func(endpoint.X)



import json
from collections.abc import Sequence
from datetime import datetime
from typing import Any, Dict, List, Optional, Type, Union

import numpy  # type: ignore
import pandas  # type: ignore
from pydantic import BaseModel, Field, conlist, validator

from tktl.core.serializers.base import CustomDeserializingModelT
from tktl.core.serializers.utils import get_list_shape


def get_single_value_model(_value: Any) -> Type[CustomDeserializingModelT]:
    class SingleValueModel(CustomDeserializingModelT):
        value: type(_value)  # type: ignore

        def deserialize(self):
            return self.value

    return SingleValueModel


def get_dataframe_model(
    base_model: Type[BaseModel], example: List[Dict], unique_id: Optional[str] = None
) -> Type[CustomDeserializingModelT]:
    validates_dt_model = _dt_validator_model(base_model, unique_id=unique_id)

    class DataFrame(CustomDeserializingModelT):
        __root__: List[validates_dt_model]  # type: ignore

        class Config:
            schema_extra = {"example": example}

        def __iter__(self):
            return iter(self.__root__)

        def deserialize(self):
            values = self.dict()["__root__"]
            index = [v.pop("index") for v in values]
            if not all(index):
                return pandas.DataFrame.from_records(values)
            else:
                return pandas.DataFrame.from_records(
                    values, index=pandas.Index(data=index)
                )

    return DataFrame


def get_series_model(
    series: pandas.Series, base_model: Type[BaseModel], example: Dict[str, List] = None
) -> Type[CustomDeserializingModelT]:
    class Series(base_model, CustomDeserializingModelT):  # type: ignore
        _name = series.name
        _dtype = series.dtype

        def deserialize(self):
            values = self.dict()[self._name]
            _series = pandas.Series(values)
            _series = _series.astype(self._dtype)  # column types
            return _series

        class Config:
            schema_extra = {"example": example}

    return Series


def get_jsonable_encoder_sequence_model(
    base_model: Type[BaseModel],
) -> Type[CustomDeserializingModelT]:
    class JsonableEncoderModel(base_model, CustomDeserializingModelT):  # type: ignore
        def deserialize(self):
            return json.loads(self.values)

    return JsonableEncoderModel


def get_nested_sequence_model(sequence: Sequence) -> Type[CustomDeserializingModelT]:
    shape, types = get_list_shape(sequence)
    shape = (None,) + shape[1:]
    inner_model = conlist(Union[types], max_items=shape[-1], min_items=shape[-1])  # type: ignore
    for s in reversed(shape[:-1]):
        inner_model = conlist(inner_model, max_items=s, min_items=s)
    return _nested_sequence_model(inner_models=inner_model)  # type: ignore


def get_array_model(base_model, example: List[Any]) -> Type[CustomDeserializingModelT]:
    class ArrayModel(base_model, CustomDeserializingModelT):  # type: ignore
        def deserialize(self):
            return numpy.array(self.dict()["__root__"])

        class Config:
            schema_extra = {"example": example}

    return ArrayModel


def get_flat_array_model(inner_type: type, example: List[Any]):
    class FlatArray(BaseModel):
        __root__: List[inner_type] = Field(..., alias="value")  # type: ignore

        def deserialize(self):
            return numpy.array(self.dict()["value"])

        class Config:
            schema_extra = {"example": example}

    return FlatArray


def get_mapping_model(base_model: Type[BaseModel]) -> Type[CustomDeserializingModelT]:
    class MappingModel(base_model, CustomDeserializingModelT):  # type: ignore
        def deserialize(self):
            return self.dict()

    return MappingModel


def _dt_validator_model(base_model, unique_id=None):
    class DtValidateModel(base_model):
        @validator("*", each_item=True, allow_reuse=True)
        def remove_tz(cls, v):  # type: ignore
            if isinstance(v, datetime):
                return v.replace(tzinfo=None)
            return v

    DtValidateModel.__name__ = unique_id
    return DtValidateModel


class SequenceEncoderModel(CustomDeserializingModelT):
    __root__: List

    def __iter__(self):
        return iter(self.__root__)

    def deserialize(self):
        return self.dict()["__root__"]


def _nested_sequence_model(inner_models: Type[BaseModel]):
    class NestedSequenceModel(SequenceEncoderModel):
        __root__: inner_models  # type: ignore

    return NestedSequenceModel


def get_custom_model(base_model) -> Type[CustomDeserializingModelT]:
    class CustomModel(base_model, CustomDeserializingModelT):  # type: ignore
        def deserialize(self):
            return self.dict()

    return CustomModel



import pytest
from pydantic import UUID4

from tktl.commands.deployments import GetDeployments, GetEndpoints
from tktl.login import logout


@pytest.mark.skip(reason="This relies on our production system. Replace me")
def test_get_deployment_commands(logged_in_context):
    cmd = GetDeployments()
    result = cmd.execute(
        UUID4("33fb5b8f-e6a2-4dd2-86ef-d257d2c59b85"),
        None,
        None,
        None,
        None,
        None,
        return_all=False,
    )
    assert len(result) == 1

    result = cmd.execute(None, None, None, None, None, None, return_all=True)
    assert len(result) >= 5

    cmd = GetEndpoints()
    result = cmd.execute(
        UUID4("427cae9a-275f-469f-acc0-5478614ec863"),
        None,
        None,
        None,
        None,
        None,
        None,
        None,
    )
    assert len(result) == 4

    cmd = GetDeployments()
    logout()
    assert not cmd.execute(
        UUID4("7c0f6f48-0220-450a-b4d2-bfc731f94cc3"),
        None,
        None,
        None,
        None,
        None,
        return_all=True,
    )



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""The setup script."""
from pathlib import Path

from setuptools import find_packages, setup  # type: ignore

with open("README.md", encoding="utf-8") as readme_file:
    readme = readme_file.read()


def strip_comments(line):
    return line.split("#", 1)[0].strip()


def _pip_requirement(req, *root):
    if req.startswith("-r "):
        _, path = req.split()
        return reqs(*root, *path.split("/"))
    return [req]


def _reqs(*f):
    path = (Path.cwd() / "reqs").joinpath(*f)
    with path.open() as fh:
        reqs = [strip_comments(line) for line in fh.readlines()]
        return [_pip_requirement(r, *f[:-1]) for r in reqs if r]


def reqs(*f):
    return [req for subreq in _reqs(*f) for req in subreq]


install_requires = reqs("base.txt")
test_requires = reqs("test.txt") + install_requires

setup(
    author="Taktile GmbH",
    author_email="devops@taktile.com",
    license="Apache License 2.0",
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
    ],
    description="Taktile's CLI ",
    python_requires=">=3.7, <3.11",
    entry_points={
        "console_scripts": [
            "tktl=tktl.main:cli",
        ],
    },
    long_description=readme,
    long_description_content_type="text/markdown",
    include_package_data=True,
    keywords="taktile-cli",
    name="taktile-cli",
    setup_requires=install_requires,
    install_requires=install_requires,
    test_suite="tests",
    packages=find_packages(),
    tests_require=test_requires,
    url="https://docs.taktile.com",
    version="1.5.6b16",
    zip_safe=False,
)



from typing import Any

import pandas  # type: ignore


def pandas_to_python_type(series: pandas.Series):
    if pandas.api.types.is_bool_dtype(series):
        return type(True)
    elif pandas.api.types.is_float_dtype(series):
        return type(1.0)
    elif pandas.api.types.is_integer_dtype(series):
        return type(1)
    elif pandas.api.types.is_string_dtype(series):
        if pandas.api.types.is_object_dtype(series):
            return Any
        else:
            return type("a")
    else:
        return None



from distutils.version import StrictVersion

from tktl import __version__
from tktl.version import TaktileVersionChecker, VersionChecker


def test_get_version(capsys):
    TaktileVersionChecker.look_for_new_version_with_timeout()
    out, err = capsys.readouterr()
    assert out == ""
    vc = VersionChecker()
    up_to_date, version_from_repository = vc.is_up_to_date("taktile-cli", __version__)
    if __version__[0:3] == version_from_repository[0:3]:
        assert StrictVersion(__version__) >= StrictVersion(
            version_from_repository
        ), f"version: {__version__}, version_from_repository: {version_from_repository}"
        assert up_to_date is True
